"""
Functions that help processing model data
"""
from struct import unpack
from pathlib import Path
import re
import os

from typing import Dict, Callable, Tuple, List, Union, IO, Optional, Set

import numpy as np
from obspy.core import Trace, Stream, read as read_ms
from obspy import UTCDateTime

from sohstationviewer.model.downsampler import downsample
from sohstationviewer.model.mseed.blockettes_reader import (
    read_next_blkt, ReadBlocketteError,
)

from sohstationviewer.conf.dbSettings import dbConf
from sohstationviewer.conf import constants as const
from sohstationviewer.view.util.enums import LogType


def read_mseed_chanids_from_headers(path2file: Path, filename: str) \
        -> Tuple[Set[str], Set[str], Set[str], Set[str]]:
    """
    from the given file get set of channel ids for soh, mass position,
        waveform, soh with sample rate greater than 1 which are
        calibration signals .
    :param path2file: path to file
    :param filename: name of file
    :return soh_chan_ids: list of chan_ids for SOH
    :return mass_pos_chan_ids: list of chan_ids for mass position
    :return wf_chan_ids: list of chan_ids for waveform
    :return spr_gr_1_chan_ids: list of chan_ids for SOH with sample rate > 1
    """
    try:
        stream = read_ms(path2file,
                         headonly=True)
    except TypeError:
        return [], [], [], []
    except Exception:
        raise Exception("Cannot determine file and block sizes. File: %s"
                        % filename)

    soh_chan_ids = set()
    mass_pos_chan_ids = set()
    wf_chan_ids = set()
    spr_gr_1_chan_ids = set()
    for trace in stream:
        chan_id = trace.stats['channel'].strip()
        chan_type = check_chan(chan_id, [], ['*'], True, True)
        if chan_type is False:
            continue
        if chan_type == 'MP':
            mass_pos_chan_ids.add(chan_id)
        elif chan_type == 'SOH':
            if trace.stats['sampling_rate'] <= 1:
                soh_chan_ids.add(chan_id)
            else:
                spr_gr_1_chan_ids.add(chan_id)
        else:
            wf_chan_ids.add(chan_id)
    return soh_chan_ids, mass_pos_chan_ids, wf_chan_ids, spr_gr_1_chan_ids


def read_mseed_or_text(
        path2file: Path, tmp_dir: str,
        read_start: Optional[float], read_end: Optional[float],
        stream_header_by_key_chan: Dict[str, Dict[str, Stream]],
        soh_data: Dict, mass_pos_data: Dict,
        waveform_data: Dict,
        log_data: Dict[str, Union[List[str], Dict[str, List[str]]]],
        data_time: Dict[str, List[float]],
        req_soh_chans: List[str], req_wf_chans: List[str],
        include_mp123zne: bool, include_mp456uvw: bool,
        track_info: Callable[[str, LogType], None],
        populate_cur_key_for_all_data: Callable[
            [Union[str, Tuple[str, str]]], None]) -> Dict[str, List[str]]:
    """
    Use read() from obspy.core to read miniseed file
       + If there is error, try to read text to add to log_data
       + If there isn't error, call read_mseed() to process
    :param path2file: absolute path of mseed file
    :param tmp_dir: dir to keep memmap files
    :param read_start: selected start epoch
    :param read_end: selected end epoch
    :param stream_header_by_key_chan: stream header by sta_id, chan_id to get
        gaps later
    :param soh_data: holder of SOH data
    :param mass_pos_data: holder of mass position data
    :param waveform_data: holder of waveform data
    :param log_data: holder of info from log
    :param data_time: time range of data sets by station id
    :param req_wf_chans: requested waveform channel list
    :param req_soh_chans: requested SOH channel list
    :param include_mp123zne: if mass position channels 1,2,3 are requested
    :param include_mp456uvw: if mass position channels 4,5,6 are requested
    :param track_info: to display processing info
    :param populate_cur_key_for_all_data: function to populate the cur_key
        to all data dicts
    """
    if read_start is not None:
        starttime = UTCDateTime(read_start)
        endtime = UTCDateTime(read_end)
    else:
        starttime = None
        endtime = None

    ret_type = check_mseed_header(
        path2file, stream_header_by_key_chan,
        soh_data, mass_pos_data, waveform_data, log_data,
        req_soh_chans, req_wf_chans, include_mp123zne, include_mp456uvw,
        populate_cur_key_for_all_data)
    if ret_type == 'NO REQ':
        return
    if ret_type == 'WRONG FORMAT':
        try:
            read_text(path2file, log_data['TEXT'])
        except Exception as e:
            track_info(str(e), LogType.WARNING)
        return

    # TODO: modify obspy to read stream according to list of indexes which
    #   can be retrieved from check_mseed_header
    #   to make perfomance faster when user set preferred channels
    #   to read and don't have to worry about treating ASCII differently
    if ret_type == 'ASCII':
        # When reading ASCII data with time obspy will skip any trace with
        # no sample which is the nature of ASCII data.
        stream = read_ms(path2file)
    else:       # HAS REQ
        stream = read_ms(path2file, starttime=starttime, endtime=endtime)

    read_mseed(
        path2file, tmp_dir, stream, starttime, endtime,
        soh_data, mass_pos_data, waveform_data, log_data,
        data_time, req_soh_chans, req_wf_chans,
        include_mp123zne, include_mp456uvw, track_info)


def check_mseed_header(
        path2file: Path,
        stream_header_by_key_chan: Dict[str, Dict[str, Stream]],
        soh_data: Dict, mass_pos_data: Dict,
        waveform_data: Dict, log_data: Dict,
        req_soh_chans: List[str], req_wf_chans: List[str],
        include_mp123zne: bool, include_mp456uvw: bool,
        populate_cur_key_for_all_data: Callable[
            [Union[str, Tuple[str, str]]], None]):
    """
    Read mseed headers of a file to check chan_type, decode of mseed file,
        initiate channels to not miss any one, even the empty one,
        when actually read in detail
    :param path2file: absolute path of mseed file
    :param starttime: start of read data to skip reading actual data if not
        in range
    :param endtime: end of read data to skip reading actual data if not
        in range
    :param stream_header_by_key_chan: dict of stream header by sta, chan to get
        gaps later
    :param req_soh_chans: requested SOH channel list
    :param req_wf_chans: requested waveform channel list
    :param include_mp123zne: if mass position channels 1,2,3 are requested
    :param include_mp456uvw: if mass position channels 4,5,6 are requested
    :param populate_cur_key_for_all_data: function to populate the cur_key
        to all data dicts

    :return:
        "WRONG FORMAT": if read() return TypeError
        "ASCII": if any required channel encoded ASCII read
        "HAS REQ": if any other required channel read
        "NO REQ": if no required channel read
    """
    try:
        stream = read_ms(path2file, headonly=True)
    except TypeError:
        return "WRONG FORMAT"
    has_req_data = False
    ret_type = None
    for trace in stream:
        chan_id = trace.stats['channel'].strip()
        chan_type = check_chan(chan_id, req_soh_chans, req_wf_chans,
                               include_mp123zne, include_mp456uvw)
        if chan_type is False:
            continue
        sta_id = trace.stats['station'].strip()
        samplerate = trace.stats['sampling_rate']

        if chan_type == 'SOH' and req_soh_chans == [] and samplerate > 1:
            # If 'All chans' is selected for SOH, channel with samplerate>1
            # will be skipped by default to improve performance.
            # Note: If user intentionally added channels with samplerate>1
            # using SOH Channel Preferences dialog, they are still read.
            continue
        populate_cur_key_for_all_data(sta_id)

        has_req_data = True
        sta_id = trace.stats['station'].strip()
        if chan_id not in stream_header_by_key_chan[sta_id]:
            stream_header_by_key_chan[sta_id][chan_id] = Stream()
        stream_header_by_key_chan[sta_id][chan_id].append(trace)

        if trace.stats.mseed['encoding'] == 'ASCII':
            ret_type = 'ASCII'
            continue

        if chan_type == 'MP':
            if chan_id not in mass_pos_data[sta_id]:
                mass_pos_data[sta_id][chan_id] = {'tracesInfo': [],
                                                  'samplerate': samplerate,
                                                  'chanID': chan_id}
        elif chan_type == 'SOH':
            if chan_id not in soh_data[sta_id]:
                soh_data[sta_id][chan_id] = {
                    'tracesInfo': [],
                    'samplerate': samplerate,
                    'chanID': chan_id}
        else:  # WF
            if chan_id not in waveform_data[sta_id]:
                waveform_data[sta_id][chan_id] = {
                    "samplerate": samplerate,
                    "tracesInfo": [],
                    'chanID': chan_id}

    if ret_type is None:
        if has_req_data:
            ret_type = 'HAS REQ'
        else:
            ret_type = 'NO REQ'
    return ret_type


def read_mseed(path2file: Path, tmp_dir: str, stream: Stream,
               starttime: UTCDateTime, endtime: UTCDateTime,
               soh_data: Dict, mass_pos_data: Dict, waveform_data: Dict,
               log_data: Dict[str, Union[List[str], Dict[str, List[str]]]],
               data_time:  Dict[str, List[float]],
               req_soh_chans: List[str], req_wf_chans: List[str],
               include_mp123zne: bool, include_mp456uvw: bool,
               track_info: Callable[[str, LogType], None]):
    """
    Read traces from the given stream to add to soh_stream and waveform_data:
       + if encoding is ASCII: log string will be added to log_data. The
           following blockettes' data if available will be converted to string
           to add to log_data as well.
       + otherwise traces from stream will be add to soh_data to be
           merged later or will be convert to memmap and add memmap's name to
           waveform_data
    :param path2file: absolute path of mseed file
    :param tmp_dir: dir to keep memmap files
    :param stream: obspy stream read from file
    :param starttime: where to start picking trace for reading
    :param endtime: where to stop picking time for reading
    :param soh_data: holder of SOH data
    :param mass_pos_data: holder of mass position data
    :param waveform_data: holder of waveform data
    :param log_data: holder of info from log
    :param data_time: time range of data sets by station id
    :param req_soh_chans: requested SOH channel list
    :param req_wf_chans: requested waveform channel list
    :param include_mp123zne: if mass position channels 1,2,3 are requested
    :param include_mp456uvw: if mass position channels 4,5,6 are requested
    :param track_info: to display processing info
    """
    file = None
    for trace in stream:
        if (starttime > trace.stats['endtime'] or
                endtime < trace.stats['starttime']):
            continue
        chan_id = trace.stats['channel'].strip()
        # If possible to fix obspy to read specific channels, don't need
        # to read all trace
        chan_type = check_chan(chan_id, req_soh_chans, req_wf_chans,
                               include_mp123zne, include_mp456uvw)
        if chan_type is False:
            continue
        sta_id = trace.stats['station'].strip()
        samplerate = trace.stats['sampling_rate']

        if trace.stats.mseed['encoding'] == 'ASCII':
            file, errmsg = read_ascii(
                path2file, file, trace, log_data[sta_id])
            if errmsg != '':
                track_info(f"{path2file.name}: {sta_id} - {chan_id}: {errmsg}",
                           LogType.ERROR)
        else:
            if chan_type == 'SOH' and req_soh_chans == [] and samplerate > 1:
                # If 'All chans' is selected for SOH, channel with samplerate>1
                # will be skipped by default to improve performance.
                # Note: If user intentionally added channels with samplerate>1
                # using SOH Channel Preferences dialog, they are still read.
                #
                # TODO: this condition can be remove after obspy is modified
                continue

            if chan_type == 'MP':
                traces_info = mass_pos_data[sta_id][chan_id]['tracesInfo']
            elif chan_type == 'SOH':
                traces_info = soh_data[sta_id][chan_id]['tracesInfo']
            else:   # WF
                traces_info = waveform_data[sta_id][chan_id][
                    "tracesInfo"]
            tr = read_mseed_trace(
                trace, sta_id, chan_id, len(traces_info), tmp_dir)
            traces_info.append(tr)

            data_time[sta_id][0] = min(
                trace.stats['starttime'].timestamp, data_time[sta_id][0])
            data_time[sta_id][1] = max(
                trace.stats['endtime'].timestamp, data_time[sta_id][1])

    if file is not None:
        file.close()


def read_mseed_trace(trace: Trace, sta_id: Union[Tuple[str, str], str],
                     chan_id: str, tr_idx: int, tmp_dir: str) -> Dict:
    """
    Read mseed trace using read_mseed_trace_spr_less_than_or_equal_1()
        for sample-rate <= 1
        and read_mseed_trace_spr_greater_than_1 for sample-rate > 1
    :param trace: mseed trace
    :param sta_id: station name
    :param chan_id: channel name
    :param tr_idx: index of trace in tracesInfo list
    :param tmp_dir: path to the directory that store memmap files
    :return dict of trace's info
    """

    if trace.stats.sampling_rate <= 1:
        return read_mseed_trace_spr_less_than_or_equal_1(trace)
    else:
        return read_mseed_trace_spr_greater_than_1(
            trace, sta_id, chan_id, tr_idx, tmp_dir)


def read_mseed_trace_spr_less_than_or_equal_1(trace: Trace) -> Dict:
    """
    For mseed trace of which sample rate <=1, read and keep all info of the
        trace including times and data in memory.
    Traces that have sample_rate <=1 can be soh, mass position, waveform

    :param trace: mseed trace
    :return tr: dict of trace's info in which data and times are kept
    """
    tr = {}
    tr['chanID'] = trace.stats.channel
    tr['startTmEpoch'] = trace.stats.starttime.timestamp
    tr['endTmEpoch'] = trace.stats.endtime.timestamp
    tr['samplerate'] = trace.stats.sampling_rate
    tr['size'] = trace.stats.npts
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    tr['times'] = trace.times() + trace.stats['starttime'].timestamp
    if trace.stats.channel.startswith('MassPos'):
        tr['data'] = convert_reftek_masspos_data(trace.data)
    else:
        tr['data'] = trace.data
    return tr


def read_mseed_trace_spr_greater_than_1(
        trace: Trace, sta_id: Union[Tuple[str, str], str],
        chan_id: str, tr_idx: int, tmp_dir: str) -> Dict:
    """
    For mseed trace with sample rate (spr) > 1, info will be read but data will
        will be saved in a file to be retrieved later when needed because
        data can be very big.
    Traces that have spr>1 can be soh, mass position, waveform.
    (For soh, traces with spr>1 only be considered when the channels are
        explicitly stated in channel preferences)

    :param trace: mseed trace
    :param sta_id: station name
    :param chan_id: channel name
    :param tr_idx: index of trace in tracesInfo list
    :param tmp_dir: path to the directory that store memmap files
    :return tr: dict of trace's info in which,
        + times: aren't kept and will be reconstruct using np.linspace
            fed with startTmEpoch, endTmEpoch and samplerate
        + data: are saved in a file of which name is kept in tr['data_f']
    """
    tr = {}
    tr['chanID'] = trace.stats.channel
    tr['samplerate'] = trace.stats.sampling_rate
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    tr['size'] = trace.stats.npts
    if trace.stats.channel.startswith('MassPos'):
        data = convert_reftek_masspos_data(trace.data)
    else:
        data = trace.data
    tr['data_f'] = save_data_2_file(
        tmp_dir, 'data', sta_id, chan_id, data, tr_idx, tr['size'])
    return tr


def convert_reftek_masspos_data(data: np.ndarray) -> Dict:
    """
    Read mass possition trace's info using read_mseed_trace_spr_lt1(), then
        calculate real value for mass possition

    :param data: mseed data
    :return data that has been converted from 16-bit signed integer in which
        32767= 2 ** 16/2 - 1 is the highest value of 16-bit two's complement
        number. The value is also multiplied by 10 for readable display.
    (According to 130_theory.pdf: Each channel connects to a 12-bit A/D
    converter with an input range of +/- 10V. These channel are read
    once per second as left-justified, 2's-compliment, 16 bit values.)

    """
    return np.round_(data / 32767.0 * 10.0, 1)


def combine_traces_except_gaps_overlaps(
        traces: List[Dict], gaps: List[List[float]]) -> List[Dict]:
    """
    Return new list of traces for a channel in which traces are combined
        together but split into different traces at gaps/overlaps.

    :param traces: list of traces of the channel
    :param gaps: list of gaps of the channel
    :return new_traces: list of new traces
    """
    end_epoch = 0
    current_times = []
    current_data = []
    start_epoch = traces[0]['startTmEpoch']
    size = 0
    new_traces = []
    curr_gap_idx = 0
    for idx, tr in enumerate(traces):
        try:
            if (min(gaps[curr_gap_idx]) <= tr['startTmEpoch']
                    <= max(gaps[curr_gap_idx]) and size != 0):
                new_tr = {'samplerate': tr['samplerate'],
                          'startTmEpoch': start_epoch,
                          'endTmEpoch': end_epoch,
                          'size': size,
                          'times': np.hstack(current_times),
                          'data': np.hstack(current_data)}
                new_traces.append(new_tr)
                curr_gap_idx += 1
                start_epoch = tr['startTmEpoch']
                current_times = []
                current_data = []
                size = 0
        except IndexError:
            pass
        current_times.append(tr['times'])
        current_data.append(tr['data'])
        size += tr['size']
        end_epoch = tr['endTmEpoch']

    new_tr = {
        'samplerate': tr['samplerate'],
        'startTmEpoch': start_epoch,
        'endTmEpoch': end_epoch,
        'size': size,
        'times': np.hstack(current_times),
        'data': np.hstack(current_data)}
    new_traces.append(new_tr)
    return new_traces


def read_ascii(path2file: Path, file: IO,
               trace: Trace, sta_log_data: dict) -> IO:
    """
    Read mseed trace with ASCII encoding to add to log_data.
    :param path2file: str- absolute path of mseed file
    :param file: file object - to continue reading. Open new if file is None
    :param chan_id: str - channel ID got from mseed header
    :param trace: obspy.core.trace.Trace - mseed trace
    :param log_data: dict - holder for log messages, refer to
        DataTypeModel.__init__.log_data
    :param track_info: function - to display processing info
    :return file: file object - to close or continue to read
    """
    errmsg = ''
    byteorder = trace.stats.mseed['byteorder']
    h = trace.stats
    log_text = "\n\nSTATE OF HEALTH: "
    log_text += f"From:{h.starttime}  To:{h.starttime}\n"
    text_from_data = trace.data.tobytes().decode()
    if text_from_data != '':
        log_text += text_from_data
    else:
        rec_len = h.mseed['record_length']
        if file is None:
            file = open(path2file, 'rb')
        data_bytes = file.read(rec_len)
        following_blkts_total = unpack('%s%s' % (byteorder, 'B'),
                                       data_bytes[39:40])[0]
        if following_blkts_total > 1:
            next_blkt_byte_no = 48 + 8  # header + blkt1000(SEED info)
        else:
            next_blkt_byte_no = 0
        while next_blkt_byte_no != 0:
            try:
                next_blkt_byte_no, info = read_next_blkt(
                    next_blkt_byte_no, data_bytes, byteorder)
                log_text += info
            except ReadBlocketteError as e:
                errmsg = e.msg
                break

    if h.channel not in sta_log_data:
        sta_log_data[h.channel] = []
    sta_log_data[h.channel].append(log_text)
    return file, errmsg


def read_text(path2file: Path, text_logs: List) -> bool:
    """
    Read text file and add to log_data under channel TEXT.
        + Raise exception if the file isn't a text file
        + Remove empty lines in content
    :param path2file: str - absolute path to text file
    :param file_name: str - name of text file
    :param text_logs: holder to keep log string, refer to
        DataTypeModel.__init__.log_data['TEXT']
    """
    def is_binary(content):
        return lambda bytes: bool(bytes.translate(None, content))

    with open(path2file, 'r') as file:
        try:
            content = file.read().strip()
        except UnicodeDecodeError:

            raise Exception("Can't process file: %s" % path2file.name)

        if is_binary(content):
            raise Exception("Skip non-mseed binary file: %s" % path2file.name)
        if content != '':
            # skip empty lines
            no_empty_line_list = [
                line for line in content.splitlines() if line]
            no_empty_line_content = os.linesep.join(no_empty_line_list)

            log_text = "\n\n** STATE OF HEALTH: %s\n" % path2file.name
            log_text += no_empty_line_content
            text_logs.append(log_text)
        return True


def save_data_2_file(tmp_dir: str, tm_data: str, sta_id: str, chan_id: str,
                     tr: np.ndarray, tr_idx: int, tr_size: int) -> Path:
    """
    Using np.memmap save time/data to file then free memory for processing
    :param tmp_dir: str - the temporary root_dir to save file in
    :param tm_data: str - "times"/"data"
    :param sta_id: str - station ID
    :param chan_id: str - channel ID
    :param tr: numpy array - of trace time or data
    :param tr_idx: int - trace index
    :param tr_size: int - trace size
    :return mem_file_name: str - name of memmap file
    """
    mem_file_name = Path(tmp_dir).joinpath(
        f"{sta_id}-{chan_id}-{tm_data}-{tr_idx}")
    mem_file = np.memmap(mem_file_name, dtype='int64', mode='w+',
                         shape=tr_size)
    mem_file[:] = tr[:]
    del tr
    del mem_file
    return mem_file_name


def check_chan(chan_id: str, req_soh_chans: List[str], req_wf_chans: List[str],
               include_mp123zne: bool, include_mp456uvw: bool) \
        -> Union[str, bool]:
    """
    Check if chanID is a requested channel.
    :param chan_id: str - channel ID
    :param req_soh_chans: list of str - requested SOH channels
    :param req_wf_chans: list of str - requested waveform channels
    :param include_mp123zne: if mass position channels 1,2,3 are requested
    :param include_mp456uvw: if mass position channels 4,5,6 are requested

    :return: str/bool -
        'WF' if chanID is a requested waveform channel,
        'SOH' if chanID is a requested SOH channel,
        'MP' if chanID is a requested mass position channel
        False otherwise.
    """
    if chan_id.startswith('VM'):
        if (not include_mp123zne and
                chan_id[-1] in ['1', '2', '3', 'Z', 'N', 'E']):
            return False
        if (not include_mp456uvw
                and chan_id[-1] in ['4', '5', '6', 'U', 'V', 'W']):
            return False
        return 'MP'

    ret = check_wf_chan(chan_id, req_wf_chans)
    if ret[0] == 'WF':
        if ret[1]:
            return "WF"
        else:
            return False
    if check_soh_chan(chan_id, req_soh_chans):
        return "SOH"
    return False


def check_soh_chan(chan_id: str, req_soh_chans: List[str]) -> bool:
    """
    Check if chan_id is a requested SOH channel.
    Mass position is always included.
    This function is used for mseed only so mass position is 'VM'.
    If there is no reqSOHChans, it means all SOH channels are requested
    :param chan_id: str - channel ID
    :param req_soh_chans: list of str - requested SOH channels
    :return: bool - True if chan_id is a requested SOH channel. False otherwise
    """
    if req_soh_chans == []:
        return True
    if chan_id in req_soh_chans:
        return True
    if 'EX?' in req_soh_chans and chan_id.startswith('EX'):
        if chan_id[2] in ['1', '2', '3']:
            return True
    # TODO: remove mass position channels from SOH
    if chan_id.startswith('VM'):
        if chan_id[2] in ['0', '1', '2', '3', '4', '5', '6']:
            return True
    return False


def check_wf_chan(chan_id: str, req_wf_chans: List[str]) -> Tuple[str, bool]:
    """
    Check if chanID is a waveform channel and is requested by user
    :param chan_id: str - channel ID
    :param req_wf_chans: list of str - requested waveform channels
    :return wf: str - '' if chan_id is not a waveform channel.
                      'WF' if chan_id is a waveform channel.
    :return has_chan: bool - True if chan_id is a requested waveform channel.
    """
    if not dbConf['seisRE'].match(chan_id):
        return '', False

    for req in req_wf_chans:
        if len(req) == 1:
            req = req.replace('*', '...')
        elif len(req) == 2:
            req = req.replace('*', '..')
        elif len(req) == 3:
            req = req.replace('*', '.')

        if re.compile(f'^{req}$').match(chan_id):
            return 'WF', True

    return 'WF', False


def sort_data(data_dict: Dict) -> None:
    """
    Sort data in 'traces_info' in 'startTmEpoch' order
    :param data_dict: DataTypeModel.__init__.waveform_data
    """
    for chan_id in data_dict:
        traces_info = data_dict[chan_id]['tracesInfo']
        data_dict[chan_id]['tracesInfo'] = sorted(
            traces_info, key=lambda i: i['startTmEpoch'])


def check_related_gaps(min1: float, max1: float,
                       min2: float, max2: float,
                       index: int, checked_indexes: List[int]):
    """
    Check if the passing ranges overlapping each other and add indexes to
        checked_indexes.
    :param min1: start of range 1
    :param max1: end of range 1
    :param min2: start of range 2
    :param max2: end of range 2
    :param index: index of gap being checked
    :param checked_indexes: list of gaps that have been checked

    :return: True if the two ranges overlap each other, False otherwise
    :rtype:
    """
    if ((min1 <= min2 <= max1) or (min1 <= max2 <= max1)
            or (min2 <= min1 <= max2) or (min2 <= max1 <= max2)):
        # range [min1, max1] and [min2, max2] have some part overlap each other
        checked_indexes.append(index)
        return True
    else:
        return False


def squash_gaps(gaps: List[List[float]]) -> List[List[float]]:
    """
    Compress gaps from different channels that have time range related to
    each other to the ones with outside boundary (min start, max end)
    or (min end, max start) in case of overlap.
    :param gaps: [[float, float],], [[float, float],] -
        list of gaps of multiple channels: [[start, end],], [[start, end],]
    :return: squashed_gaps: [[float, float],] - all related gaps are squashed
        extending to the outside start and end
        [[min start, max end], [max start, min end]]

    """
    gaps = sorted(gaps, key=lambda x: x[0])
    squashed_gaps = []
    checked_indexes = []

    for idx, g in enumerate(gaps):
        if idx in checked_indexes:
            continue
        squashed_gaps.append(g)
        checked_indexes.append(idx)
        overlap = g[0] >= g[1]
        for idx_, g_ in enumerate(gaps):
            if idx_ in checked_indexes:
                continue
            if not overlap:
                if g_[0] >= g_[1]:
                    continue
                if check_related_gaps(g[0], g[1], g_[0], g_[1],
                                      idx_, checked_indexes):
                    squashed_gaps[-1][0] = min(g[0], g_[0])
                    squashed_gaps[-1][1] = max(g[1], g_[1])
                else:
                    break
            else:
                if g_[0] < g_[1]:
                    continue
                if check_related_gaps(g[1], g[0], g_[1], g_[0],
                                      idx_, checked_indexes):
                    squashed_gaps[-1][0] = max(g[0], g_[0])
                    squashed_gaps[-1][1] = min(g[1], g_[1])

    return squashed_gaps


def trim_downsample_chan_with_spr_less_or_equal_1(chan, start_tm, end_tm):
    """
    When channel is zoomed in, times and data will be trimmed off to exclude
        the non-included time. Then they will be downsampled to get to the
        expected const.CHAN_SIZE_LIMIT, and save to chan[times], chan[data]
        for plotting.

    :param chan: dict - channel info from soh_data, refer
        DataType_Model.__init_.soh_data[key][chan_id]
    :param start_tm: float - start time of zoomed section
    :param end_tm: float - end time of zoomed section
    """
    # zoom into the given time
    times = []
    data = []
    log_idx = []
    for tr in chan['tracesInfo']:
        indexes = np.where(
            (start_tm <= tr['times']) & (tr['times'] <= end_tm))[0]
        if len(indexes) == 0:
            continue
        if 'logIdx' in tr.keys():
            tr_times, tr_data, tr_logidx = downsample(
                tr['times'][indexes], tr['data'][indexes],
                tr['logIdx'][indexes],
                rq_points=const.CHAN_SIZE_LIMIT)
            log_idx.append(tr_logidx)
        else:
            tr_times, tr_data, _ = downsample(
                tr['times'][indexes], tr['data'][indexes],
                rq_points=const.CHAN_SIZE_LIMIT)
        times.append(tr_times)
        data.append(tr_data)
    chan['times'] = times
    chan['data'] = data
    if 'reftek' in chan:
        chan['logIdx'] = log_idx


def trim_waveform_data(wf_channel_data: Dict, start_time: float,
                       end_time: float) -> List[Dict]:
    """
    Trim off waveform traces whose times does not intersect the closed interval
    [start_time, end_time].

    :param wf_channel_data: dict - a dictionary that contains the waveform
        information of a channel. For its structure, refer to
        DataTypeModel.waveform_data[key]['readData'][chan_id].
    :param start_time: the start of the time range to trim waveform data to.
    :param end_time: the end of the time range to trim waveform data to.
    :return: a list with all the waveform traces that were not trimmed.
    """
    if ((start_time > wf_channel_data['tracesInfo'][-1]['endTmEpoch']) or
            (end_time < wf_channel_data['tracesInfo'][0]['startTmEpoch'])):
        return []

    good_start_indices = [index
                          for index, tr
                          in enumerate(wf_channel_data['tracesInfo'])
                          if tr['startTmEpoch'] > start_time]
    if good_start_indices:
        start_idx = good_start_indices[0]
        if start_idx > 0:
            start_idx -= 1  # start_time in middle of trace
    else:
        start_idx = 0

    good_end_indices = [idx
                        for idx, tr
                        in enumerate(wf_channel_data['tracesInfo'])
                        if tr['endTmEpoch'] <= end_time]
    if good_end_indices:
        end_idx = good_end_indices[-1]
        if end_idx < len(wf_channel_data['tracesInfo']) - 1:
            end_idx += 1  # end_time in middle of trace
    else:
        end_idx = 0
    end_idx += 1  # a[x:y+1] = [a[x], ...a[y]

    trimmed_traces_list = wf_channel_data['tracesInfo'][start_idx:end_idx]

    return trimmed_traces_list


def downsample_waveform_data(trimmed_traces_list: List[Dict],
                             start_time: float, end_time: float
                             ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Downsample trimmed waveform data. Works on one trace at a time instead
    of combining the traces before downsampling.

    :param trimmed_traces_list: a list of traces that have been trimmed to
        include only those that intersect the closed interval
        [start_time, end_time]
    :param start_time: the start of the time range
    :param end_time: the end of the time range
    :return: list of downsampled times array and list of downsampled data array
    """
    # Calculate the number of requested_points
    total_size = sum([tr['size'] for tr in trimmed_traces_list])
    requested_points = 0
    if total_size > const.CHAN_SIZE_LIMIT:
        requested_points = int(
            const.CHAN_SIZE_LIMIT / len(trimmed_traces_list)
        )

    # Downsample the data
    downsampled_times_list = []
    downsampled_data_list = []
    for tr_idx, tr in enumerate(trimmed_traces_list):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        indexes = np.where((start_time <= times) & (times <= end_time))
        times = times[indexes]
        data = data[indexes]
        if requested_points != 0:
            times, data, _ = downsample(times, data,
                                        rq_points=requested_points)
        downsampled_times_list.append(times)
        downsampled_data_list.append(data)

    return downsampled_times_list, downsampled_data_list


def trim_downsample_wf_chan(chan: Dict, start_tm: float, end_tm: float,
                            firsttime: bool) -> Optional[bool]:
    """
    When channel is zoomed in, times and data will be trimmed off to exclude
        the non-included time.
    Different with soh_data where times and data are each in one np.array,
        in waveform_data, times and data are each kept in a list of np.memmap
        files along with startTmEpoch and endTmEpoch.
    Data in np.memmap files will be trimmed again if there is still time
        outside the zoom time range then downsampled and combined to get the
        times and data with size closed to the expected const.CHAN_SIZE_LIMIT,
        and save to chan[times], chan[data] for plotting.
    chan['fulldata']: is a flag set at the firsttime when channel isn't
        zoomed in and the data is small enough so it doesn't need to go through
        the process when zoomed in.

    :param chan: dict - channel info from soh_data, refer
        DataType_Model.__init_.soh_data[key][chan_id]
    :param start_tm: float - start time of zoomed section
    :param end_tm: float - end time of zoomed section
    :param firsttime: bool True for original size when channel is not zoomed in
    """
    if 'fulldata' in chan:
        # data is small, already has full in the first trim
        return

    z_traces_info = trim_waveform_data(chan, start_tm, end_tm)
    if not z_traces_info:
        return False

    total_size = sum([tr['size'] for tr in z_traces_info])
    if not firsttime and total_size > const.RECAL_SIZE_LIMIT:
        # size too big, recalc only when it's small enough
        return
    if total_size <= const.CHAN_SIZE_LIMIT and firsttime:
        chan['fulldata'] = True

    try:
        del chan['times']
        del chan['data']
    except Exception:
        pass

    downsampled_times, downsampled_data = downsample_waveform_data(
        z_traces_info, start_tm, end_tm
    )

    chan['times'] = np.hstack(downsampled_times)
    chan['data'] = np.hstack(downsampled_data)


def get_start_5mins_of_diff_days(start_tm: float, end_tm: float) -> np.ndarray:
    """
    Get the list of the start time of all five minutes for each day start from
        the day of startTm and end at the day of endTm.
    :param start_tm: float - start time
    :param end_tm: float - end time
    :return start_5mins_of_diff_days: [[288 of floats], ] - the list of
        start of all five minutes of days specified by start_tm and end_tm in
        which each day has 288 of 5 minutes.
    """
    exact_day_tm = (start_tm // const.SEC_DAY) * const.SEC_DAY
    exact_day_tm_list = []

    if start_tm < exact_day_tm:
        exact_day_tm_list = [exact_day_tm - const.SEC_DAY]

    while exact_day_tm < end_tm:
        exact_day_tm_list.append(exact_day_tm)
        exact_day_tm += const.SEC_DAY

    # list of start/end 5m in each day: start_5mins_of_diff_days
    for idx, start_day_tm in enumerate(exact_day_tm_list):
        start_5mins_of_day = np.arange(start_day_tm,
                                       start_day_tm + const.SEC_DAY,
                                       const.SEC_5M)
        if idx == 0:
            start_5mins_of_diff_days = np.array([start_5mins_of_day])
        else:
            start_5mins_of_diff_days = np.vstack(
                (start_5mins_of_diff_days, start_5mins_of_day))
    return start_5mins_of_diff_days


def find_tps_tm_idx(
        given_tm: float, start_5mins_of_diff_days: List[List[float]]) \
        -> Tuple[float, float]:
    """
    Find the position of the given time (given_tm) in time-power-squared plot
    :param given_tm: float - given time
    :param start_5mins_of_diff_days: [[288 of floats], ] - the list of
        start of all five minutes of some specific days in which each day has
        288 of 5 minutes.
    :return x_idx: int - index of 5m section
    :return y_idx: int - index of the day the given time belong to in plotting
    """
    x_idx = None
    y_idx = None
    for day_idx, a_day_5mins in enumerate(start_5mins_of_diff_days):
        for start_5m_idx, start_5m in enumerate(a_day_5mins):
            if start_5m > given_tm:
                # index of day start from 0 to negative because day is plotted
                # from top to bottom
                y_idx = - day_idx
                x_idx = start_5m_idx - 1
                if start_5m_idx == 0:
                    # if the start_5m_idx == 0, the given time belong to the
                    # last 5m of the previous day
                    y_idx = -(day_idx - 1)
                    x_idx = const.NO_5M_DAY - 1
                break
        if x_idx is not None:
            break

    if x_idx is None:
        # x_idx == None happens when the given time fall into the last 5m of
        # the last day. Although the time 24:00 of the last day belongs
        # to the next days of other cases, but since there is no more days to
        # plot it, it is no harm to set it at the last 5m of the last day.
        x_idx = const.NO_5M_DAY - 1
        y_idx = - (len(start_5mins_of_diff_days) - 1)

    return x_idx, y_idx


def retrieve_gaps_from_stream_header(
        streams: Dict[str, Dict[str, Stream]],
        gaps_by_key_chan: Dict[Union[str, Tuple[str, str]],
                               Dict[str, List[List[int]]]],
        gaps: Dict[str, List[List[float]]],
        read_start: Optional[float], read_end: Optional[float]) -> \
        Dict[str, List[List[float]]]:
    """
    Retrieve gaps by sta_id from stream_header_by_key_chan

    :param streams: dict of stream header by sta, chan
    :param gaps_by_key_chan: gaps list by key and channel id
    :param gaps: gaps list by key
    """
    for sta_id in streams:
        sta_gaps = []
        gaps_by_key_chan[sta_id] = {}
        for chan_id in streams[sta_id]:
            stream = streams[sta_id][chan_id]
            gaps_in_stream = stream.get_gaps()
            gaps_by_key_chan[sta_id][chan_id] = stream_gaps = [
                [g[4].timestamp, g[5].timestamp] for g in gaps_in_stream
                if read_start <= min(g[4].timestamp, g[5].timestamp) <= read_end or  # noqa
                   read_start <= max(g[4].timestamp, g[5].timestamp) <= read_end]    # noqa

            sta_gaps += stream_gaps
        gaps[sta_id] = squash_gaps(sta_gaps)


if __name__ == '__main__':
    print(check_chan('HH1', [], ['*']))
